import os
import time
import sys
import torch
from tensorboardX import SummaryWriter
import subprocess

ROOT_PROJECT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
ROOT_DATA_PATH = os.path.join(ROOT_PROJECT_PATH, "data")


class Recorder:
    def __init__(self,
                 experiment_name,
                 data_dir_name,
                 ):
        print('Recorder init')
        self.experiment_name = experiment_name
        self.data_dir_name = data_dir_name
        self.data_path = os.path.join(ROOT_DATA_PATH, data_dir_name)
        self._create_dir(self.data_path)
        self.title_reward = {}

    def _create_dir(self, path):
        """
        创建tensorboard结果文件的储存文件夹
        :param path:
        :return:
        """
        if not os.path.exists(path):
            os.makedirs(path)
        print('Recorder 数据存储文件夹创建完成：' + path)
        return path

    def create_data_result(self, title):
        self.title_reward.update(
            {title: {'reward_list': []}}
        )

    def get_data_path(self):
        return self.data_path

    def add_reward(self, title, reward):
        self.title_reward[title]['reward_list'].append(reward)

    def get_reward_list(self, title):
        return self.title_reward[title]['reward_list']

    def writeAll2TensorBoard(self):
        """
        :param title: str,数据标题
        :return:
        """
        writer = SummaryWriter(log_dir=self.data_path)
        for title in self.title_reward.keys():
            for i in range(len(self.title_reward[title]['reward_list'])):
                writer.add_scalar(tag=title, scalar_value=self.title_reward[title]['reward_list'][i], global_step=i)
        writer.close()

    def openTensorBoard(self):
        """
        :param path: 当前工程路径
        :param directory: tensorboard文件的存储目录名称
        :return:
        """
        cmds = [ROOT_PROJECT_PATH.split("\\")[0], "cd " + ROOT_PROJECT_PATH, "tensorboard --logdir=data"]
        p = subprocess.Popen('cmd.exe', stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for cmd in cmds:
            p.stdin.write((cmd + "\n").encode(encoding='UTF-8', errors='strict'))
        p.stdin.close()
        print(p.stdout.read())

    def create_save_model(self, agent):
        def save_model(model_names):
            for name in model_names:
                torch.save(agent.__dict__[name].state_dict(), self.data_path + '/' + name + '.pt')
                print('Trainify 模型 ' + name + ' 保存完毕')

        return save_model

    def create_load_model(self, agent):
        def load_model(model_names):
            for name in model_names:
                agent.__dict__[name].load_state_dict(torch.load(self.data_path + '/' + name + '.pt'))
                print('Trainify 模型 ' + name + ' 加载完毕')

        return load_model


if __name__ == '__main__':
    a = Recorder('ttt', 'ttt_results')
    a.create_data_result("test1")
    for i in range(100):
        a.add_reward("test1", i * 6)
    a.create_data_result("test2")
    for i in range(100):
        a.add_reward("test2", i * 10)
    a.writeAll2TensorBoard()
    a.openTensorBoard()



import matplotlib.pyplot as plt

def evaluate_pendulum(agent,env):
    reward_list2 = []
    min_reward = 0
    g_cos_l = 1

    for l in range(10):
        reward = 0
        s0 = env.reset()
        cos_l = 1
        for step in range(10000):
            # envs.render()
            a0 = agent.act(s0)
            s1, r1, done, _ = env.step(a0)
            if s1[0] < cos_l:
                cos_l = s1[0]
            if s1[0] <= 0:
                print('fall down-------', s1)
            reward += r1
            s0 = s1
        if cos_l < g_cos_l:
            g_cos_l = cos_l
        print('evaluate: ', l, ':', reward, cos_l, g_cos_l)

        if reward < min_reward:
            min_reward = reward
        reward_list2.append(reward)
    plt.plot(reward_list2)
    plt.show()
    return min_reward



from torch.utils.tensorboard import SummaryWriter
import subprocess


def writeTensorBoard(directory, title, data):
    """
    :param directory: str,tensorboard文件的存储目录
    :param title: str,数据标题
    :param data: 一维列表,要写到tensorboard中的数据
    :return:
    """
    writer = SummaryWriter(log_dir= directory)
    for i in range(len(data)):
        writer.add_scalar(tag= title, scalar_value= data[i], global_step=i)
    writer.close()

def openTensorBoard(path, directory):
    """
    :param path: 当前工程路径
    :param directory: tensorboard文件的存储目录名称
    :return:
    """
    cmds = [path.split("\\")[0], "cd "+path, "tensorboard --logdir="+directory]
    p = subprocess.Popen('cmd.exe', stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    for cmd in cmds:
        p.stdin.write((cmd + "\n").encode(encoding='UTF-8', errors='strict'))
    p.stdin.close()
    print(p.stdout.read())


import os
import sys
import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from trainify.utils import str_to_list

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)


class Actor(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super(Actor, self).__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear1.weight.data.normal_(0, 0.1)
        self.linear1.bias.data.zero_()
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear2.weight.data.normal_(0, 0.1)
        self.linear2.bias.data.zero_()
        self.linear3 = nn.Linear(hidden_size, output_size)
        self.linear3.weight.data.normal_(0, 0.1)
        self.linear3.bias.data.zero_()

    def forward(self, s):
        x = torch.tanh(self.linear1(s))
        x = torch.tanh(self.linear2(x))
        x = self.linear3(x)

        return x


class Critic(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super().__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear3 = nn.Linear(hidden_size, output_size)

    def forward(self, s, a):
        x = torch.cat([s, a], 1)
        x = F.relu(self.linear1(x))
        x = F.relu(self.linear2(x))
        x = self.linear3(x)
        return x


class DDPGAgent(object):
    # def __init__(self, divide_tool, env):
    def __init__(self, env, config):
        self.config = config
        self.originEnv = env
        self.env = env
        self.gamma = config['gamma']
        self.actor_lr = config['actor_lr']
        self.critic_lr = config['critic_lr']
        self.tau = config['tau']
        self.capacity = config['capacity']
        self.batch_size = config['batch_size']
        self.hidden_size = config['hidden_size']
        s_dim = self.env.observation_space.shape[0] * 2
        a_dim = self.env.action_space.shape[0]

        self.actor = Actor(s_dim, self.hidden_size, a_dim)
        self.network = Actor(s_dim, self.hidden_size, a_dim)
        self.actor_target = Actor(s_dim, self.hidden_size, a_dim)
        self.critic = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.critic_target = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.actor_optim = optim.Adam(self.actor.parameters(), lr=self.actor_lr)
        self.critic_optim = optim.Adam(self.critic.parameters(), lr=self.critic_lr)
        self.buffer = []

        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target.load_state_dict(self.critic.state_dict())

    def reset(self):
        self.env = self.originEnv
        self.gamma = self.config['gamma']
        self.actor_lr = self.config['actor_lr']
        self.critic_lr = self.config['critic_lr']
        self.tau = self.config['tau']
        self.capacity = self.config['capacity']
        self.batch_size = self.config['batch_size']

        s_dim = self.originEnv.observation_space.shape[0] * 2
        a_dim = self.originEnv.action_space.shape[0]

        self.actor = Actor(s_dim, self.hidden_size, a_dim)
        self.network = Actor(s_dim, self.hidden_size, a_dim)
        self.actor_target = Actor(s_dim, self.hidden_size, a_dim)
        self.critic = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.critic_target = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.actor_optim = optim.Adam(self.actor.parameters(), lr=self.actor_lr)
        self.critic_optim = optim.Adam(self.critic.parameters(), lr=self.critic_lr)
        self.buffer = []
        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target.load_state_dict(self.critic.state_dict())

    def act(self, s0):
        abs = str_to_list(s0)
        # abs = str_to_list(self.divide_tool.get_abstract_state(s0))
        # s0 = torch.tensor(s0, dtype=torch.float).unsqueeze(0)
        s0 = torch.tensor(abs, dtype=torch.float).unsqueeze(0)
        a0 = self.actor(s0).squeeze(0).detach().numpy()
        return a0

    def put(self, *transition):
        if len(self.buffer) == self.capacity:
            self.buffer.pop(0)
        self.buffer.append(transition)

    def learn(self):
        if len(self.buffer) < self.batch_size:
            return

        samples = random.sample(self.buffer, self.batch_size)

        s0, a0, r1, s1 = zip(*samples)

        s0 = torch.tensor(np.array(s0), dtype=torch.float)
        a0 = torch.tensor(np.array(a0), dtype=torch.float)
        r1 = torch.tensor(np.array(r1), dtype=torch.float).view(self.batch_size, -1)
        s1 = torch.tensor(np.array(s1), dtype=torch.float)

        def critic_learn():
            a1 = self.actor_target(s1).detach()
            y_true = r1 + self.gamma * self.critic_target(s1, a1).detach()

            y_pred = self.critic(s0, a0)

            loss_fn = nn.MSELoss()
            loss = loss_fn(y_pred, y_true)
            self.critic_optim.zero_grad()
            loss.backward()
            self.critic_optim.step()

        def actor_learn():
            loss = -torch.mean(self.critic(s0, self.actor(s0)))
            self.actor_optim.zero_grad()
            loss.backward()
            self.actor_optim.step()

        def soft_update(net_target, net, tau):
            for target_param, param in zip(net_target.parameters(), net.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        critic_learn()
        actor_learn()
        soft_update(self.critic_target, self.critic, self.tau)
        soft_update(self.actor_target, self.actor, self.tau)



import time

import numpy as np
from rtree import index

from trainify.validator import FMValidator

record_num = 1


def cegar(rtree_name, agent, divide_tool, train_model, verify_env, config):
    max_iteration = config.max_iteration

    print('Verify cegar 开始时间' + time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))
    t0 = time.time()
    # 此时agent应该已经训练好了
    # train_model(agent)
    agent.load()
    # evaluate(agent)
    tr = time.time()
    v = FMValidator(verify_env, agent.network)
    k = v.create_kripke_ctl()
    t1 = time.time()
    if k is None:
        violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
        res2 = True
        print('Verify cegar number of counterexamples：', len(violated_states))
    else:
        # v.formula = 'not(A(G(safe)))'
        res2, violated_states = v.ctl_model_check(k)
        print('Verify cegar number of counterexamples：', len(violated_states))
    t2 = time.time()
    print('Verify cegar train time:', tr - t0, 'construct kripke structure:', t1 - tr, 'model checking:', t2 - t1)
    print(time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))
    start_id = divide_tool.rtree.get_size()
    p = index.Property()
    p.dimension = len(divide_tool.key_dim)
    iteration_time = 1
    while res2 and iteration_time <= max_iteration:
        print(time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))
        t0 = time.time()
        divide_tool.rtree = index.Index(rtree_name, divide_tool.rtree_refinement(violated_states, start_id),
                                        properties=p)
        print('Verify cegar number of states after refinement:', divide_tool.rtree.get_size())
        trr = time.time()
        start_id += divide_tool.rtree.get_size()
        agent.divide_tool = divide_tool
        # agent.load()
        agent.reset()
        train_model(agent)

        agent.load()
        # evaluate(agent)
        tr = time.time()
        verify_env.divide_tool = divide_tool
        verify_env.network = agent.network
        # pd = PendulumEnv(divide_tool, agent.actor)
        v = FMValidator(verify_env, agent.network)
        k = v.create_kripke_ctl()
        t1 = time.time()
        if k is None:
            violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
            res2 = True
            print('Verify cegar number of counterexamples：', len(violated_states))
        else:
            # v.formula = 'not(A(G(safe)))'
            res2, violated_states = v.ctl_model_check(k)
            print('Verify cegar number of counterexamples：', len(violated_states))
        print('Verify cegar iteration :', iteration_time, res2, len(violated_states))
        t2 = time.time()
        print('Verify cegar refine:', trr - t0, 'train:', tr - trr, 'construct kripke structure:', t1 - tr,
              'model checking:',
              t2 - t1)
        print(time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))
        iteration_time += 1
    while res2:
        t0 = time.time()
        # agent.load()
        agent.reset()
        train_model(agent)

        agent.load()
        tr = time.time()
        # verify_env.divide_tool = divide_tool
        verify_env.network = agent.network
        v = FMValidator(verify_env, agent.network)
        k = v.create_kripke_ctl()
        t1 = time.time()
        if k is None:
            violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
            res2 = True
            print('Verify cegar number of counterexamples：', len(violated_states))
        else:
            # v.formula = 'not(A(G(safe)))'
            res2, violated_states = v.ctl_model_check(k)
            print('Verify cegar number of counterexamples：', len(violated_states))
        t2 = time.time()
        print('Verify cegar train:', tr - t0, 'construct kripke structure:', t1 - tr, 'model checking:', t2 - t1)
        print(time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Trainify-proto",
    version="0.0.4",
    author="ericx",
    author_email="ericxlee@formail.com",
    description="Trainify-proto",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jieye-ericx/Trainify-proto",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'torch==1.10.0',
        'pandas==1.3.4',
        'psutil==5.8.0',
        'Rtree==0.9.7',
        'lark==0.11.3',
        'gym==0.23.0',
        'matplotlib==3.4.3',
        'numpy==1.20.0',
        'scipy==1.8.1',
    ]
)
